library(Biostrings)
library(parallel)

construct_hvr_sequence <- function(data) {
  vdj_dataframe <- as.data.frame(data)
  hvr_sequence <- paste(vdj_dataframe$v_sequence,
    vdj_dataframe$d_sequence,
    vdj_dataframe$j_sequence,
    sep = ""
  )
  vdj_dataframe$hvr <- hvr_sequence
  return(vdj_dataframe)
}

parse_data <- function(files) {
  reversed_sequences <- Biostrings::readQualityScaledDNAStringSet(files[1])
  sequences <- Biostrings::reverseComplement(reversed_sequences)
  vdj_alignment <- read.csv(files[2])
  vdj_dataframe <- construct_hvr_sequence(vdj_alignment)
  vj_segments <- union(
    readRDS("data/v_segments.rds"),
    readRDS("data/j_segments_phe.rds")
  )
  return(list(sequences, vj_segments, vdj_dataframe))
}

# TODO Test if grep can return more than one match
get_vj_sequence <- function(identifier, names, sequences) {
  row <- grep(names, pattern = identifier)
  return(as.character(sequences[row]))
}

construct_full_sequences <- function(vdj_segments, metadata) {
  v_sequences <- lapply(metadata$v_call,
    names(vdj_segments),
    vdj_segments,
    FUN = get_vj_sequence
  )
  j_sequences <- lapply(metadata$j_call,
    names(vdj_segments),
    vdj_segments,
    FUN = get_vj_sequence
  )
  full_sequence <- paste(v_sequences, metadata$hvr, j_sequences, sep = "")
  return(Biostrings::DNAStringSet(full_sequence))
}

align_sequence <- function(sequence, vdj_segment) {
  return(Biostrings::pairwiseAlignment(
    pattern = sequence,
    subject = vdj_segment,
    type = "global-local",
    gapOpening = 1,
  ))
}

perform_alignment <- function(sequences, vdj_segments, metadata) {
  vj_sequences <- construct_full_sequences(vdj_segments, metadata)
  sequence_alignment <- mcmapply(vj_sequences,
    vdj_segments,
    FUN = align_sequence,
    mc.cores = detectCores()
  )
  return(sequence_alignment)
}

input_files <- c("data/curesim_sequence.fastq", "data/vdj_alignment.csv")
data <- parse_data(input_files)
alignment <- perform_alignment(
  sequences = data[[1]],
  vdj_segments = data[[2]],
  metadata = data[[3]]
)
print(alignment)
